"""
Module containing exceptions thrown by :mod:`dxf`.
"""

class DXFError(Exception):
    """
    Base exception class for all dxf errors
    """

class DXFUnexpectedError(DXFError):
    """
    Unexpected value error
    """
    def __init__(self, got, expected):
        """
        :param got: Actual value received
        :param expected: Value that was expected
        """
        super(DXFUnexpectedError, self).__init__()
        self.got = got
        self.expected = expected

class DXFUnexpectedStatusCodeError(DXFUnexpectedError):
    """
    Unexpected HTTP status code
    """
    def __str__(self):
        return 'expected status code %d, got %d' % (self.expected, self.got)

class DXFDigestMismatchError(DXFUnexpectedError):
    """
    Digest didn't match expected value
    """
    def __str__(self):
        return 'expected digest %s, got %s' % (self.expected, self.got)

class DXFUnexpectedKeyTypeError(DXFUnexpectedError):
    """
    Cryptographic key type not supported
    """
    def __str__(self):
        return 'expected key type %s, got %s' % (self.expected, self.got)

class DXFUnexpectedDigestMethodError(DXFUnexpectedError):
    """
    Digest method not supported
    """
    def __str__(self):
        return 'expected digest method %s, got %s' % (self.expected, self.got)

class DXFDisallowedSignatureAlgorithmError(DXFError):
    """
    Signature algorithm forbidden
    """
    def __init__(self, alg):
        """
        :param alg: Forbidden signature algorithm
        :type alg: str
        """
        super(DXFDisallowedSignatureAlgorithmError, self).__init__()
        self.alg = alg

    def __str__(self):
        return 'disallowed signature algorithm: %s' % self.alg

class DXFSignatureChainNotImplementedError(DXFError):
    """
    Signature chains not supported
    """
    def __str__(self):
        return 'verification with a cert chain is not implemented'

class DXFUnauthorizedError(DXFError):
    """
    Registry returned authorized error
    """
    def __str__(self):
        return 'unauthorized'

class DXFAuthInsecureError(DXFError):
    """
    Can't authenticate over insecure (non-HTTPS) connection
    """
    def __str__(self):
        return 'Auth requires HTTPS'

class DXFDigestNotAvailableForSchema1(DXFError):
    """
    https://github.com/docker/distribution/issues/1662#issuecomment-213101772
    A schema1 manifest should always produce the same image id but defining the
    steps to produce directly from the manifest is not straight forward."
    """
    def __str__(self):
        return 'Schema 1 manifest has no configuration blob. See https://github.com/docker/distribution/issues/1662#issuecomment-213101772'

class DXFMountFailed(DXFError):
    """
    Failed to mount blob from another repository.
    """
    def __str__(self):
        return 'Cross repository blob mount failed'



import os
import sys
import errno
import time
import hashlib
import requests.exceptions
import pytest
import tqdm
import dxf.main

# pylint: disable=no-member

def test_empty(dxf_main, capsys):
    assert dxf.main.doit(['list-repos'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == ""
    assert err == ""

def _not_found(dxf_main, name):
    assert dxf.main.doit(['blob-size', pytest.repo, name], dxf_main) == errno.ENOENT

def test_not_found(dxf_main):
    _not_found(dxf_main, pytest.blob1_hash)
    _not_found(dxf_main, pytest.blob2_hash)
    _not_found(dxf_main, '@fooey')

def test_push_blob(dxf_main, capsys):
    assert dxf.main.doit(['push-blob', pytest.repo, pytest.blob1_file], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['push-blob', pytest.repo, pytest.blob2_file], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob2_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['get-alias', pytest.repo, 'fooey'], dxf_main) == errno.ENOENT
    out, err = capsys.readouterr()
    assert out == ""
    assert err.index('Not Found') >= 0
    assert dxf.main.doit(['push-blob', pytest.repo, pytest.blob1_file, '@fooey'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['get-alias', pytest.repo, 'fooey'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['list-repos'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.repo + os.linesep
    assert err == ""

def _pull_blob(dxf_main, name, dgst, capfdbinary, repo=None):
    assert dxf.main.doit(['pull-blob', pytest.repo if repo is None else repo, name], dxf_main) == 0
    out, err = capfdbinary.readouterr()
    sha256 = hashlib.sha256()
    sha256.update(out)
    assert 'sha256:' + sha256.hexdigest() == dgst
    assert err == b""

def test_pull_blob(dxf_main, capfdbinary):
    environ = {'DXF_BLOB_INFO': '1'}
    environ.update(dxf_main)
    assert dxf.main.doit(['pull-blob', pytest.repo, pytest.blob1_hash, pytest.blob2_hash], environ) == 0
    out, err = capfdbinary.readouterr()
    out_sha256 = hashlib.sha256()
    out_sha256.update(out)
    expected_sha256 = hashlib.sha256()
    expected_sha256.update(pytest.blob1_hash.encode('utf-8'))
    expected_sha256.update(b' ')
    expected_sha256.update(str(pytest.blob1_size).encode('utf-8'))
    expected_sha256.update(os.linesep.encode('utf-8'))
    with open(pytest.blob1_file, 'rb') as f:
        for chunk in iter(lambda: f.read(8192), b''):
            expected_sha256.update(chunk)
    expected_sha256.update(pytest.blob2_hash.encode('utf-8'))
    expected_sha256.update(b' ')
    expected_sha256.update(str(pytest.blob2_size).encode('utf-8'))
    expected_sha256.update(os.linesep.encode('utf-8'))
    with open(pytest.blob2_file, 'rb') as f:
        for chunk in iter(lambda: f.read(8192), b''):
            expected_sha256.update(chunk)
    assert out_sha256.digest() == expected_sha256.digest()
    assert err == b""
    _pull_blob(dxf_main, pytest.blob1_hash, pytest.blob1_hash, capfdbinary)
    _pull_blob(dxf_main, pytest.blob2_hash, pytest.blob2_hash, capfdbinary)
    _pull_blob(dxf_main, '@fooey', pytest.blob1_hash, capfdbinary)

def test_progress(dxf_main, capfd):
    environ = {'DXF_PROGRESS': '1'}
    environ.update(dxf_main)
    assert dxf.main.doit(['pull-blob', pytest.repo, pytest.blob1_hash], environ) == 0
    _, err = capfd.readouterr()
    assert pytest.blob1_hash[0:8] in err
    assert " 0%" in err
    assert " 100%" in err
    assert " " + str(pytest.blob1_size) + "/" + str(pytest.blob1_size) in err
    assert dxf.main.doit(['push-blob', pytest.repo, pytest.blob3_file], environ) == 0
    _, err = capfd.readouterr()
    assert pytest.blob3_hash[0:8] in err
    assert " 0%" in err
    assert " 100%" in err
    assert " " + str(pytest.blob3_size) + "/" + str(pytest.blob3_size) in err

def test_see_progress(dxf_main, monkeypatch):
    environ = {'DXF_PROGRESS': '1'}
    environ.update(dxf_main)
    # pylint: disable=too-few-public-methods
    class FakeStdout(object):
        # pylint: disable=no-self-use
        def write(self, _):
            time.sleep(0.05)
        def flush(self):
            pass
    monkeypatch.setattr(sys, 'stdout', FakeStdout())
    assert dxf.main.doit(['pull-blob', pytest.repo, pytest.blob1_hash], environ) == 0
    orig_tqdm = tqdm.tqdm
    def new_tqdm(*args, **kwargs):
        tqdm_obj = orig_tqdm(*args, **kwargs)
        class TQDM(object):
            # pylint: disable=no-self-use
            def update(self, n):
                tqdm_obj.update(n)
                time.sleep(0.025)
            def close(self):
                tqdm_obj.close()
            @property
            def n(self):
                return tqdm_obj.n
            @property
            def total(self):
                return tqdm_obj.total
        return TQDM()
    monkeypatch.setattr(tqdm, 'tqdm', new_tqdm)
    assert dxf.main.doit(['push-blob', pytest.repo, pytest.blob4_file], environ) == 0

def test_set_alias(dxf_main, capsys):
    assert dxf.main.doit(['set-alias', pytest.repo, 'hello', pytest.blob1_hash], dxf_main) == 0
    _, err = capsys.readouterr()
    assert err == ""
    if dxf_main['REGVER'] != 2.2:
        assert dxf.main.doit(['del-alias', pytest.repo, 'hello'], dxf_main) == 0
        out, err = capsys.readouterr()
        assert out == pytest.blob1_hash + os.linesep
        assert err == ""
        # Deleting tag actually deletes by DCD:
        # https://github.com/docker/distribution/issues/1566
        # So fooey gets deleted too
        assert dxf.main.doit(['list-aliases', pytest.repo], dxf_main) == 0
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""
        assert dxf.main.doit(['set-alias', pytest.repo, 'hello', pytest.blob1_hash], dxf_main) == 0
        assert dxf.main.doit(['set-alias', pytest.repo, 'fooey', pytest.blob1_hash], dxf_main) == 0
        _, err = capsys.readouterr()
        assert err == ""
    assert dxf.main.doit(['set-alias', pytest.repo, 'there', pytest.blob1_hash, pytest.blob2_hash], dxf_main) == 0
    _, err = capsys.readouterr()
    assert err == ""
    assert dxf.main.doit(['set-alias', pytest.repo, 'world', pytest.blob2_file], dxf_main) == 0
    _, err = capsys.readouterr()
    assert err == ""

def test_get_alias(dxf_main, capsys):
    assert dxf.main.doit(['get-alias', pytest.repo, 'hello'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['get-alias', pytest.repo, 'there'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep + \
                  pytest.blob2_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['get-alias', pytest.repo, 'world'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob2_hash + os.linesep
    assert err == ""

def test_get_digest(dxf_main, capsys):
    if dxf_main['REGVER'] == 2.2:
        with pytest.raises(dxf.exceptions.DXFDigestNotAvailableForSchema1):
            dxf.main.doit(['get-digest', pytest.repo, 'hello'], dxf_main)
        return
    assert dxf.main.doit(['get-digest', pytest.repo, 'hello'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['get-digest', pytest.repo, 'there'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob1_hash + os.linesep
    assert err == ""
    assert dxf.main.doit(['get-digest', pytest.repo, 'world'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob2_hash + os.linesep
    assert err == ""
    pytest.copy_registry_image(dxf_main['REGVER'])
    assert dxf.main.doit(['get-digest',
                          'test/registry',
                          str(dxf_main['REGVER'])],
                         dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == dxf_main['REG_DIGEST'] + os.linesep
    assert err == ""

def test_blob_size(dxf_main, capsys):
    assert dxf.main.doit(['blob-size', pytest.repo, pytest.blob1_hash, pytest.blob2_hash, '@hello', '@there', '@world'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == str(pytest.blob1_size) + os.linesep + \
                  str(pytest.blob2_size) + os.linesep + \
                  str(pytest.blob1_size) + os.linesep + \
                  str(pytest.blob1_size + pytest.blob2_size) + os.linesep + \
                  str(pytest.blob2_size) + os.linesep
    assert err == ""

_def_hash = 'sha256:cb8379ac2098aa165029e3938a51da0bcecfc008fd6795f401178647f96c5b34'

def test_mount_blob(dxf_main, capfdbinary):
    if dxf_main['REGVER'] == 2.2:
        with pytest.raises(dxf.exceptions.DXFMountFailed):
            dxf.main.doit(['mount-blob', 'some/other', pytest.repo, pytest.blob1_hash], dxf_main)
    else:
        assert dxf.main.doit(['mount-blob', 'some/other', pytest.repo, pytest.blob1_hash], dxf_main) == 0
        out, err = capfdbinary.readouterr()
        assert out == pytest.blob1_hash.encode('utf-8') + os.linesep.encode('utf-8')
        assert err == b""
        _pull_blob(dxf_main, pytest.blob1_hash, pytest.blob1_hash, capfdbinary, 'some/other')
    with pytest.raises(dxf.exceptions.DXFMountFailed):
        dxf.main.doit(['mount-blob', 'some/other', pytest.repo, _def_hash], dxf_main)
    with pytest.raises(dxf.exceptions.DXFMountFailed):
        dxf.main.doit(['mount-blob', 'some/other', 'another/repo', pytest.blob1_hash], dxf_main)
    if dxf_main['REGVER'] == 2.2:
        with pytest.raises(dxf.exceptions.DXFMountFailed):
            dxf.main.doit(['mount-blob', 'some/other', pytest.repo, pytest.blob2_hash, '@blob2-mounted'], dxf_main)
    else:
        assert dxf.main.doit(['mount-blob', 'some/other', pytest.repo, pytest.blob2_hash, '@blob2-mounted'], dxf_main) == 0
        out, err = capfdbinary.readouterr()
        assert out == pytest.blob2_hash.encode('utf-8') + os.linesep.encode('utf-8')
        assert err == b""
        _pull_blob(dxf_main, '@blob2-mounted', pytest.blob2_hash, capfdbinary, 'some/other')

def test_list_aliases(dxf_main, capsys):
    assert dxf.main.doit(['list-aliases', pytest.repo], dxf_main) == 0
    out, err = capsys.readouterr()
    assert sorted(out.split(os.linesep)) == ['', 'fooey', 'hello', 'there', 'world']
    assert err == ""

def test_manifest(dxf_main, capfdbinary, monkeypatch):
    assert dxf.main.doit(['set-alias', pytest.repo, 'mani_test', pytest.blob1_hash], dxf_main) == 0
    manifest, err = capfdbinary.readouterr()
    assert manifest
    assert err == b""
    # pylint: disable=too-few-public-methods
    class FakeStdin(object):
        # pylint: disable=no-self-use
        def read(self):
            return manifest.decode()
    monkeypatch.setattr(sys, 'stdin', FakeStdin())
    assert dxf.main.doit(['get-alias', pytest.repo], dxf_main) == 0
    out, err = capfdbinary.readouterr()
    assert out.decode() == pytest.blob1_hash + os.linesep
    assert err == b""
    assert dxf.main.doit(['blob-size', pytest.repo], dxf_main) == 0
    out, err = capfdbinary.readouterr()
    assert out.decode() == str(pytest.blob1_size) + os.linesep
    assert err == b""
    assert dxf.main.doit(['pull-blob', pytest.repo], dxf_main) == 0
    out, err = capfdbinary.readouterr()
    sha256 = hashlib.sha256()
    sha256.update(out)
    assert 'sha256:' + sha256.hexdigest() == pytest.blob1_hash
    assert err == b""
    assert dxf.main.doit(['del-blob', pytest.repo], dxf_main) == 0
    assert dxf.main.doit(['pull-blob', pytest.repo], dxf_main) == errno.ENOENT

#@pytest.mark.onlytest
def test_auth(dxf_main, capsys):
    if (not dxf_main['TEST_DO_AUTH']) or (not dxf_main['TEST_DO_TOKEN']):
        assert dxf.main.doit(['auth', pytest.repo], dxf_main) == 0
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""
    else:
        assert dxf.main.doit(['auth', pytest.repo, '*'], dxf_main) == 0
        token, err = capsys.readouterr()
        assert token
        assert err == ""
        environ = {}
        environ.update(dxf_main)
        environ.pop('DXF_USERNAME', None)
        environ.pop('DXF_PASSWORD', None)
        environ.pop('DXF_AUTHORIZATION', None)
        assert dxf.main.doit(['list-repos'], environ) == 0
        out, err = capsys.readouterr()
        expected = [pytest.repo]
        if dxf_main['REGVER'] != 2.2:
            expected += ['test/registry', 'some/other']
        assert sorted(out.rstrip().split(os.linesep)) == sorted(expected)
        assert err == ""
        assert dxf.main.doit(['list-aliases', pytest.repo], environ) == errno.EACCES
        out, err = capsys.readouterr()
        assert out == ""
        environ['DXF_TOKEN'] = token.strip()
        assert dxf.main.doit(['list-aliases', pytest.repo], environ) == 0
        out, err = capsys.readouterr()
        assert sorted(out.split(os.linesep)) == ['', 'fooey', 'hello', 'mani_test', 'there', 'world']
        assert err == ""

def test_del_blob(dxf_main, capfdbinary):
    _pull_blob(dxf_main, pytest.blob2_hash, pytest.blob2_hash, capfdbinary)
    assert dxf.main.doit(['del-blob', pytest.repo, pytest.blob2_hash], dxf_main) == 0
    _not_found(dxf_main, pytest.blob2_hash)
    assert dxf.main.doit(['del-blob', pytest.repo, pytest.blob2_hash], dxf_main) == errno.ENOENT

def test_del_alias(dxf_main, capsys):
    assert dxf.main.doit(['get-alias', pytest.repo, 'world'], dxf_main) == 0
    out, err = capsys.readouterr()
    assert out == pytest.blob2_hash + os.linesep
    assert err == ""
    if dxf_main['REGVER'] == 2.2:
        with pytest.raises(requests.exceptions.HTTPError) as ex:
            dxf.main.doit(['del-alias', pytest.repo, 'world'], dxf_main)
        assert ex.value.response.status_code == requests.codes.method_not_allowed
        assert dxf.main.doit(['get-alias', pytest.repo, 'world'], dxf_main) == 0
    else:
        assert dxf.main.doit(['del-alias', pytest.repo, 'world'], dxf_main) == 0
        out, err = capsys.readouterr()
        assert out == pytest.blob2_hash + os.linesep
        # Note: test gc but it isn't needed to make a 404
        pytest.gc()
        assert dxf.main.doit(['get-alias', pytest.repo, 'world'], dxf_main) == errno.ENOENT
        assert dxf.main.doit(['del-alias', pytest.repo, 'world'], dxf_main) == errno.ENOENT

def _num_args(dxf_main, op, minimum, maximum, capsys):
    if minimum is not None:
        with pytest.raises(SystemExit):
            dxf.main.doit([op, pytest.repo] + ['a'] * (minimum - 1), dxf_main)
        out, err = capsys.readouterr()
        assert out == ""
        assert "too few arguments" in err
    if maximum is not None:
        with pytest.raises(SystemExit):
            dxf.main.doit([op, pytest.repo] + ['a'] * (maximum + 1), dxf_main)
        out, err = capsys.readouterr()
        assert out == ""
        assert "too many arguments" in err

def test_bad_args(dxf_main, capsys):
    _num_args(dxf_main, 'push-blob', 1, 2, capsys)
    _num_args(dxf_main, 'mount-blob', 2, 3, capsys)
    _num_args(dxf_main, 'set-alias', 2, None, capsys)
    _num_args(dxf_main, 'list-aliases', None, 0, capsys)
    with pytest.raises(SystemExit):
        dxf.main.doit(['push-blob', pytest.repo, pytest.blob1_file, 'fooey'], dxf_main)
    out, err = capsys.readouterr()
    assert out == ""
    assert "invalid alias" in err
    with pytest.raises(SystemExit):
        dxf.main.doit(['mount-blob', 'some/other', pytest.repo, pytest.blob2_hash, 'blob2-mounted'], dxf_main)
    out, err = capsys.readouterr()
    assert out == ""
    assert "invalid alias" in err

def test_auth_host(dxf_main):
    if dxf_main['TEST_DO_TOKEN']:
        environ = {
            'DXF_AUTH_HOST': 'localhost:5002'
        }
        environ.update(dxf_main)
        with pytest.raises(requests.exceptions.ConnectionError):
            dxf.main.doit(['list-repos'], environ)

def test_tlsverify(dxf_main):
    if dxf_main['DXF_INSECURE'] == '0':
        v = os.environ['REQUESTS_CA_BUNDLE']
        del os.environ['REQUESTS_CA_BUNDLE']
        try:
            if dxf_main['DXF_SKIPTLSVERIFY'] == '0':
                with pytest.raises(requests.exceptions.SSLError):
                    dxf.main.doit(['list-repos'], dxf_main)
            else:
                assert dxf.main.doit(['list-repos'], dxf_main) == 0
        finally:
            os.environ['REQUESTS_CA_BUNDLE'] = v

def test_tlsverify_str(dxf_main):
    if dxf_main['DXF_INSECURE'] == '0':
        v = os.environ['REQUESTS_CA_BUNDLE']
        del os.environ['REQUESTS_CA_BUNDLE']
        skip = dxf_main['DXF_SKIPTLSVERIFY']
        dxf_main['DXF_SKIPTLSVERIFY'] = '0'
        dxf_main['DXF_TLSVERIFY'] = v
        try:
            assert dxf.main.doit(['list-repos'], dxf_main) == 0
        finally:
            os.environ['REQUESTS_CA_BUNDLE'] = v
            dxf_main['DXF_SKIPTLSVERIFY'] = skip
            del dxf_main['DXF_TLSVERIFY']



import os
import hashlib
import json
import requests
import pytest
from jwcrypto import jws
import dxf.exceptions

# pylint: disable=no-member

def _not_found(dxf_obj, dgst):
    with pytest.raises(requests.exceptions.HTTPError) as ex:
        dxf_obj.blob_size(dgst)
    assert ex.value.response.status_code == requests.codes.not_found

def test_not_found(dxf_obj):
    _not_found(dxf_obj, pytest.blob1_hash)
    _not_found(dxf_obj, pytest.blob2_hash)

def test_api_version_check(dxf_obj):
    dxf_obj.api_version_check()

def test_push_blob(dxf_obj):
    assert dxf_obj.push_blob(pytest.blob1_file) == pytest.blob1_hash
    state = {
        'called': False,
        'total': 0
    }
    # pylint: disable=unused-argument
    def progress1(dgst, chunk, size):
        state['called'] = True
    assert dxf_obj.push_blob(pytest.blob1_file, progress=progress1) == pytest.blob1_hash
    assert not state['called']
    def progress2(dgst, chunk, size):
        assert size == pytest.blob2_size
        state['total'] += len(chunk)
    assert dxf_obj.push_blob(pytest.blob2_file, progress=progress2) == pytest.blob2_hash
    assert state['total'] == pytest.blob2_size
    assert dxf_obj.list_repos() == [pytest.repo]

def test_blob_size(dxf_obj):
    assert dxf_obj.blob_size(pytest.blob1_hash) == pytest.blob1_size
    assert dxf_obj.blob_size(pytest.blob2_hash) == pytest.blob2_size

def test_mount_blob(dxf_obj):
    # pylint: disable=protected-access
    dxf_obj2 = dxf.DXF(dxf_obj._host, 'some/other', dxf_obj._auth, dxf_obj._insecure, None, dxf_obj._tlsverify)
    if dxf_obj.regver == 2.2:
        with pytest.raises(dxf.exceptions.DXFMountFailed):
            dxf_obj2.mount_blob(dxf_obj._repo, pytest.blob1_hash)
    else:
        assert dxf_obj2.mount_blob(dxf_obj._repo, pytest.blob1_hash) == pytest.blob1_hash
        _pull_blob(dxf_obj2, pytest.blob1_hash, pytest.blob1_size, None)
    with pytest.raises(dxf.exceptions.DXFMountFailed):
        dxf_obj2.mount_blob(dxf_obj._repo, _def_hash)
    with pytest.raises(dxf.exceptions.DXFMountFailed):
        dxf_obj2.mount_blob('another/repo', pytest.blob1_hash)

def _pull_blob(dxf_obj, dgst, expected_size, chunk_size):
    if expected_size is None:
        it = dxf_obj.pull_blob(dgst, chunk_size=chunk_size)
    else:
        it, size = dxf_obj.pull_blob(dgst, size=True, chunk_size=chunk_size)
        assert size == expected_size
    sha256 = hashlib.sha256()
    for chunk in it:
        if chunk_size is None:
            assert len(chunk) == 8192
        else:
            assert len(chunk) == chunk_size
        sha256.update(chunk)
    assert 'sha256:' + sha256.hexdigest() == dgst

def test_pull_blob(dxf_obj):
    _pull_blob(dxf_obj, pytest.blob1_hash, None, None)
    _pull_blob(dxf_obj, pytest.blob2_hash, pytest.blob2_size, None)
    _pull_blob(dxf_obj, pytest.blob1_hash, None, 4096)
    with pytest.raises(dxf.exceptions.DXFDigestMismatchError) as ex:
        class DummySHA256(object):
            # pylint: disable=no-self-use
            def update(self, chunk):
                pass
            def hexdigest(self):
                return orig_sha256().hexdigest()
        orig_sha256 = hashlib.sha256
        hashlib.sha256 = DummySHA256
        try:
            for _ in dxf_obj.pull_blob(pytest.blob1_hash):
                pass
        finally:
            hashlib.sha256 = orig_sha256
    assert ex.value.got == 'sha256:' + hashlib.sha256().hexdigest()
    assert ex.value.expected == pytest.blob1_hash

def test_pull_and_push_blob(dxf_obj):
    it = dxf_obj.pull_blob(pytest.blob1_hash)
    state = {'total': 0}
    sha256 = hashlib.sha256()
    def progress(dgst, chunk):
        assert dgst == pytest.blob1_hash
        state['total'] += len(chunk)
        sha256.update(chunk)
    assert dxf_obj.push_blob(data=it,
                             digest=pytest.blob1_hash,
                             progress=progress,
                             check_exists=False) == \
           pytest.blob1_hash
    assert state['total'] == pytest.blob1_size
    assert 'sha256:' + sha256.hexdigest() == pytest.blob1_hash
    _pull_blob(dxf_obj, pytest.blob1_hash, pytest.blob1_size, None)

def test_set_alias(dxf_obj):
    dxf_obj.set_alias('hello', pytest.blob1_hash)
    if dxf_obj.regver != 2.2:
        assert dxf_obj.del_alias('hello') == [pytest.blob1_hash]
        assert dxf_obj.list_aliases() == []
        dxf_obj.set_alias('hello', pytest.blob1_hash)
    dxf_obj.set_alias('there', pytest.blob1_hash, pytest.blob2_hash)
    dxf_obj.set_alias('world', pytest.blob2_hash)

def test_get_alias(dxf_obj):
    assert dxf_obj.get_alias('hello') == [pytest.blob1_hash]
    assert dxf_obj.get_alias('there') == [pytest.blob1_hash, pytest.blob2_hash]
    assert dxf_obj.get_alias('world') == [pytest.blob2_hash]

def test_get_digest(dxf_obj):
    if dxf_obj.regver == 2.2:
        with pytest.raises(dxf.exceptions.DXFDigestNotAvailableForSchema1):
            dxf_obj.get_digest('hello')
        return
    assert dxf_obj.get_digest('hello') == pytest.blob1_hash
    assert dxf_obj.get_digest('there') == pytest.blob1_hash
    assert dxf_obj.get_digest('world') == pytest.blob2_hash
    pytest.copy_registry_image(dxf_obj.regver)
    # pylint: disable=protected-access
    dxf_obj2 = dxf.DXF(dxf_obj._host, 'test/registry', dxf_obj._auth, dxf_obj._insecure, None, dxf_obj._tlsverify)
    assert dxf_obj2.get_digest(str(dxf_obj.regver)) == dxf_obj.reg_digest

def test_list_aliases(dxf_obj):
    assert sorted(dxf_obj.list_aliases()) == ['hello', 'there', 'world']
    assert sorted(list(dxf_obj.list_aliases(batch_size=2, iterate=True))) == ['hello', 'there', 'world']

def test_context_manager(dxf_obj):
    with dxf_obj as odxf:
        test_list_aliases(odxf)

def test_manifest(dxf_obj):
    manifest = dxf_obj.set_alias('mani_test', pytest.blob1_hash)
    assert manifest
    if dxf_obj.regver != 2.2:
        assert dxf_obj.get_manifest('mani_test') == manifest
    #assert json.dumps(json.loads(dxf_obj.get_manifest('mani_test')),
    #                  sort_keys=True) == \
    #       json.dumps(json.loads(manifest), sort_keys=True)
    assert dxf_obj.get_alias(manifest=manifest) == [pytest.blob1_hash]
    if json.loads(manifest)['schemaVersion'] == 1:
        with pytest.raises(jws.InvalidJWSSignature):
            dxf_obj.get_alias(manifest=' '+manifest)
    if dxf_obj.regver != 2.2:
        dxf_obj.set_manifest('mani_test2', manifest)
        assert dxf_obj.get_alias('mani_test2') == [pytest.blob1_hash]

def test_unsigned_manifest_v1(dxf_obj):
    manifest = dxf_obj.make_unsigned_manifest('mani_test3', pytest.blob2_hash)
    assert manifest
    with pytest.raises(KeyError):
        dxf_obj.get_alias(manifest=manifest)
    assert dxf_obj.get_alias(manifest=manifest, verify=False) == [pytest.blob2_hash]

def test_unsigned_manifest_v2(dxf_obj):
    manifest = dxf_obj.make_manifest(pytest.blob2_hash)
    assert manifest
    assert dxf_obj.get_alias(manifest=manifest) == [pytest.blob2_hash]

#@pytest.mark.onlytest
def test_auth(dxf_obj):
    # pylint: disable=protected-access
    if not dxf_obj.test_do_auth:
        assert dxf_obj.authenticate() is None
    elif not dxf_obj.test_do_token:
        assert dxf_obj.authenticate(pytest.username, pytest.password) is None
    else:
        assert dxf_obj.authenticate(pytest.username, pytest.password, '*') == dxf_obj.token
        assert dxf_obj.token

def test_del_blob(dxf_obj):
    _pull_blob(dxf_obj, pytest.blob2_hash, None, None)
    dxf_obj.del_blob(pytest.blob2_hash)
    _not_found(dxf_obj, pytest.blob2_hash)
    with pytest.raises(requests.exceptions.HTTPError) as ex:
        dxf_obj.del_blob(pytest.blob2_hash)
    assert ex.value.response.status_code == requests.codes.not_found

def test_del_alias(dxf_obj):
    assert dxf_obj.get_alias('world') == [pytest.blob2_hash]
    if dxf_obj.regver == 2.2:
        with pytest.raises(requests.exceptions.HTTPError) as ex:
            dxf_obj.del_alias('world')
        assert ex.value.response.status_code == requests.codes.method_not_allowed
        assert dxf_obj.get_alias('world') == [pytest.blob2_hash]
    else:
        assert dxf_obj.del_alias('world') == [pytest.blob2_hash]
        # Note: test gc but it isn't needed to make a 404
        pytest.gc()
        with pytest.raises(requests.exceptions.HTTPError) as ex:
            dxf_obj.get_alias('world')
        assert ex.value.response.status_code == requests.codes.not_found
        with pytest.raises(requests.exceptions.HTTPError) as ex:
            dxf_obj.del_alias('world')
        assert ex.value.response.status_code == requests.codes.not_found

_abc_hash = 'sha256:ba7816bf8f01cfea414140de5dae2223b00361a396177a9cb410ff61f20015ad'
_def_hash = 'sha256:cb8379ac2098aa165029e3938a51da0bcecfc008fd6795f401178647f96c5b34'

def test_hash_bytes():
    assert dxf.hash_bytes(b'abc') == _abc_hash

def test_tlsverify(dxf_obj):
    # pylint: disable=protected-access
    if not dxf_obj._insecure:
        v = os.environ['REQUESTS_CA_BUNDLE']
        del os.environ['REQUESTS_CA_BUNDLE']
        try:
            if dxf_obj._tlsverify:
                with pytest.raises(requests.exceptions.SSLError):
                    dxf_obj.list_repos()
            else:
                expected = [pytest.repo]
                if dxf_obj.regver != 2.2:
                    expected += ['test/registry', 'some/other']
                assert sorted(dxf_obj.list_repos()) == sorted(expected)
        finally:
            os.environ['REQUESTS_CA_BUNDLE'] = v

def test_tlsverify_str(dxf_obj):
    # pylint: disable=protected-access
    if not dxf_obj._insecure:
        v = os.environ['REQUESTS_CA_BUNDLE']
        del os.environ['REQUESTS_CA_BUNDLE']
        tlsv = dxf_obj._tlsverify
        dxf_obj._tlsverify = v
        try:
            expected = [pytest.repo]
            if dxf_obj.regver != 2.2:
                expected += ['test/registry', 'some/other']
            assert sorted(dxf_obj.list_repos()) == sorted(expected)
        finally:
            os.environ['REQUESTS_CA_BUNDLE'] = v
            dxf_obj._tlsverify = tlsv

def test_pagination(dxf_obj):
    # pylint: disable=protected-access
    num = 11
    for i in range(num):
        name = 'test/{0}'.format(i)
        dxf_obj2 = dxf.DXF(dxf_obj._host, name, dxf_obj._auth, dxf_obj._insecure, None, dxf_obj._tlsverify)
        assert dxf_obj2.push_blob(data=b'abc', digest=_abc_hash) == _abc_hash
    expected = [pytest.repo] + ['test/{0}'.format(i) for i in range(num)]
    if dxf_obj.regver != 2.2:
        expected += ['test/registry', 'some/other']
    assert sorted(dxf_obj2.list_repos(batch_size=3)) == sorted(expected)



"""
Module for accessing a Docker v2 Registry
"""

import base64
import hashlib
import json
import sys
import warnings

try:
    import urllib.parse as urlparse
    from urllib.parse import urlencode
except ImportError:
    # pylint: disable=import-error,no-name-in-module,wrong-import-order
    from urllib import urlencode
    import urlparse

import requests
try:
    from requests.packages import urllib3
except ImportError:
    import urllib3
from urllib3.exceptions import InsecureRequestWarning

from jwcrypto import jwk, jws
import www_authenticate
# pylint: disable=wildcard-import
from dxf import exceptions

_schema2_mimetype = 'application/vnd.docker.distribution.manifest.v2+json'

if sys.version_info < (3, 0):
    _binary_type = str
else:
    _binary_type = bytes
    # pylint: disable=redefined-builtin
    long = int

def _to_bytes_2and3(s):
    return s if isinstance(s, _binary_type) else s.encode('utf-8')

def hash_bytes(buf):
    """
    Hash bytes using the same method the registry uses (currently SHA-256).

    :param buf: Bytes to hash
    :type buf: binary str

    :rtype: str
    :returns: Hex-encoded hash of file's content (prefixed by ``sha256:``)
    """
    sha256 = hashlib.sha256()
    sha256.update(buf)
    return 'sha256:' + sha256.hexdigest()

def hash_file(filename):
    """
    Hash a file using the same method the registry uses (currently SHA-256).

    :param filename: Name of file to hash
    :type filename: str

    :rtype: str
    :returns: Hex-encoded hash of file's content (prefixed by ``sha256:``)
    """
    sha256 = hashlib.sha256()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(8192), b''):
            sha256.update(chunk)
    return 'sha256:' + sha256.hexdigest()

def _raise_for_status(r):
    # pylint: disable=no-member
    if r.status_code == requests.codes.unauthorized:
        raise exceptions.DXFUnauthorizedError()
    r.raise_for_status()

def split_digest(s):
    method, digest = s.split(':')
    if method != 'sha256':
        raise exceptions.DXFUnexpectedDigestMethodError(method, 'sha256')
    return method, digest

class _ReportingFile(object):
    def __init__(self, dgst, f, cb):
        self._dgst = dgst
        self._f = f
        self._cb = cb
        self._size = requests.utils.super_len(f)
        cb(dgst, b'', self._size)
    # define __iter__ so requests thinks we're a stream
    # (models.py, PreparedRequest.prepare_body)
    # pylint: disable=non-iterator-returned
    def __iter__(self):
        assert not "called"
    # define fileno, tell and mode so requests can find length
    # (utils.py, super_len)
    def fileno(self):
        return self._f.fileno()
    def tell(self):
        return self._f.tell()
    @property
    def mode(self):
        return self._f.mode
    def read(self, n):
        chunk = self._f.read(n)
        if chunk:
            self._cb(self._dgst, chunk, self._size)
        return chunk

class _ReportingChunks(object):
    # pylint: disable=too-few-public-methods
    def __init__(self, dgst, data, cb):
        self._dgst = dgst
        self._data = data
        self._cb = cb
    def __iter__(self):
        for chunk in self._data:
            if chunk:
                self._cb(self._dgst, chunk)
            yield chunk

class PaginatingResponse(object):
    # pylint: disable=too-few-public-methods
    def __init__(self, dxf_obj, req_meth, path, header, **kwargs):
        self._meth = getattr(dxf_obj, req_meth)
        self._path = path
        self._header = header
        self._kwargs = kwargs
    def __iter__(self):
        while self._path:
            response = self._meth('get', self._path, **self._kwargs)
            self._kwargs = {}
            for v in response.json()[self._header] or []:
                yield v
            nxt = response.links.get('next')
            self._path = nxt['url'] if nxt else None

def _ignore_warnings(obj):
    # pylint: disable=protected-access
    if obj._tlsverify is False:
        warnings.filterwarnings('ignore', category=InsecureRequestWarning)

class DXFBase(object):
    # pylint: disable=too-many-instance-attributes
    """
    Class for communicating with a Docker v2 registry.
    Contains only operations which aren't related to repositories.

    Can act as a context manager. For each context entered, a new
    `requests.Session <http://docs.python-requests.org/en/latest/user/advanced/#session-objects>`_
    is obtained. Connections to the same host are shared by the session.
    When the context exits, all the session's connections are closed.

    If you don't use :class:`DXFBase` as a context manager, each request
    uses an ephemeral session. If you don't read all the data from an iterator
    returned by :meth:`DXF.pull_blob` then the underlying connection won't be
    closed until Python garbage collects the iterator.
    """
    def __init__(self, host, auth=None, insecure=False, auth_host=None, tlsverify=True):
        # pylint: disable=too-many-arguments
        """
        :param host: Host name of registry. Can contain port numbers. e.g. ``registry-1.docker.io``, ``localhost:5000``.
        :type host: str

        :param auth: Authentication function to be called whenever authentication to the registry is required. Receives the :class:`DXFBase` object and a HTTP response object. It should call :meth:`authenticate` with a username, password and ``response`` before it returns.
        :type auth: function(dxf_obj, response)

        :param insecure: Use HTTP instead of HTTPS (which is the default) when connecting to the registry.
        :type insecure: bool

        :param auth_host: Host to use for token authentication. If set, overrides host returned by then registry.
        :type auth_host: str

        :param tlsverify: When set to False, do not verify TLS certificate. When pointed to a `<ca bundle>.crt` file use this for TLS verification. See `requests.verify <http://docs.python-requests.org/en/latest/user/advanced/#ssl-cert-verification>`_ for more details.
        :type tlsverify: bool or str
        """
        self._base_url = ('http' if insecure else 'https') + '://' + host + '/v2/'
        self._host = host
        self._auth = auth
        self._insecure = insecure
        self._auth_host = auth_host
        self._token = None
        self._headers = {}
        self._repo = None
        self._sessions = [requests]
        self._tlsverify = tlsverify

    @property
    def token(self):
        """
        str: Authentication token. This will be obtained automatically when
        you call :meth:`authenticate`. If you've obtained a token
        previously, you can also set it but be aware tokens expire quickly.
        """
        return self._token

    @token.setter
    def token(self, value):
        self._token = value
        self._headers = {
            'Authorization': 'Bearer ' + value
        }

    def _base_request(self, method, path, **kwargs):
        def make_kwargs():
            r = {'allow_redirects': True, 'verify': self._tlsverify}
            r.update(kwargs)
            if 'headers' not in r:
                r['headers'] = {}
            r['headers'].update(self._headers)
            return r
        url = urlparse.urljoin(self._base_url, path)
        with warnings.catch_warnings():
            _ignore_warnings(self)
            r = getattr(self._sessions[0], method)(url, **make_kwargs())
        # pylint: disable=no-member
        if r.status_code == requests.codes.unauthorized and self._auth:
            headers = self._headers
            self._auth(self, r)
            if self._headers != headers:
                with warnings.catch_warnings():
                    _ignore_warnings(self)
                    r = getattr(self._sessions[0], method)(url, **make_kwargs())
        _raise_for_status(r)
        return r

    def authenticate(self,
                     username=None, password=None,
                     actions=None, response=None,
                     authorization=None,
                     user_agent='Docker-Client/19.03.2 (linux)'):
        # pylint: disable=too-many-arguments,too-many-locals,too-many-branches
        """
        Authenticate to the registry using a username and password,
        an authorization header or otherwise as the anonymous user.

        :param username: User name to authenticate as.
        :type username: str

        :param password: User's password.
        :type password: str

        :param actions: If you know which types of operation you need to make on the registry, specify them here. Valid actions are ``pull``, ``push`` and ``*``.
        :type actions: list

        :param response: When the ``auth`` function you passed to :class:`DXFBase`'s constructor is called, it is passed a HTTP response object. Pass it back to :meth:`authenticate` to have it automatically detect which actions are required.
        :type response: requests.Response

        :param authorization: ``Authorization`` header value.
        :type authorization: str

        :param user_agent: ``User-Agent`` header value.
        :type user_agent: str

        :rtype: str
        :returns: Authentication token, if the registry supports bearer tokens. Otherwise ``None``, and HTTP Basic auth is used (if the registry requires authentication).
        """
        if response is None:
            with warnings.catch_warnings():
                _ignore_warnings(self)
                response = self._sessions[0].get(self._base_url, verify=self._tlsverify)

        if response.ok:
            return None

        # pylint: disable=no-member
        if response.status_code != requests.codes.unauthorized:
            raise exceptions.DXFUnexpectedStatusCodeError(response.status_code,
                                                          requests.codes.unauthorized)

        if self._insecure:
            raise exceptions.DXFAuthInsecureError()

        parsed = www_authenticate.parse(response.headers['www-authenticate'])

        if username is not None and password is not None:
            headers = {
                'Authorization': 'Basic ' + base64.b64encode(_to_bytes_2and3(username + ':' + password)).decode('utf-8')
            }
        elif authorization is not None:
            headers = {
                'Authorization': authorization
            }
        else:
            headers = {}
        headers["User-Agent"] = user_agent

        if 'bearer' in parsed:
            info = parsed['bearer']
            if actions and self._repo:
                scope = 'repository:' + self._repo + ':' + ','.join(actions)
            elif 'scope' in info:
                scope = info['scope']
            elif not self._repo:
                # Issue #28: gcr.io doesn't return scope for non-repo requests
                scope = 'registry:catalog:*'
            else:
                scope = ''
            url_parts = list(urlparse.urlparse(info['realm']))
            query = urlparse.parse_qs(url_parts[4])
            query.update({
                'service': info['service'],
                'scope': scope
            })
            url_parts[4] = urlencode(query, True)
            url_parts[0] = 'https'
            if self._auth_host:
                url_parts[1] = self._auth_host
            auth_url = urlparse.urlunparse(url_parts)
            with warnings.catch_warnings():
                _ignore_warnings(self)
                r = self._sessions[0].get(auth_url, headers=headers, verify=self._tlsverify)
            _raise_for_status(r)
            rjson = r.json()
            # Use 'access_token' value if present and not empty, else 'token' value.
            self.token = rjson.get('access_token') or rjson['token']
            return self._token

        self._headers = headers
        return None

    def list_repos(self, batch_size=None, iterate=False):
        """
        List all repositories in the registry.

        :param batch_size: Number of repository names to ask the server for at a time.
        :type batch_size: int

        :param iterate: Whether to return iterator over the names or a list of all the names.
        :type iterate: bool

        :rtype: list or iterator of strings
        :returns: Repository names.
        """
        it = PaginatingResponse(self, '_base_request',
                                '_catalog', 'repositories',
                                params={'n': batch_size})
        return it if iterate else list(it)

    def __enter__(self):
        assert self._sessions
        session = requests.Session()
        session.__enter__()
        self._sessions.insert(0, session)
        return self

    def __exit__(self, *args):
        assert len(self._sessions) > 1
        session = self._sessions.pop(0)
        return session.__exit__(*args)

class DXF(DXFBase):
    """
    Class for operating on a Docker v2 repositories.
    """
    def __init__(self, host, repo, auth=None, insecure=False, auth_host=None, tlsverify=True):
        # pylint: disable=too-many-arguments
        """
        :param host: Host name of registry. Can contain port numbers. e.g. ``registry-1.docker.io``, ``localhost:5000``.
        :type host: str

        :param repo: Name of the repository to access on the registry. Typically this is of the form ``username/reponame`` but for your own registries you don't actually have to stick to that.
        :type repo: str

        :param auth: Authentication function to be called whenever authentication to the registry is required. Receives the :class:`DXF` object and a HTTP response object. It should call :meth:`DXFBase.authenticate` with a username, password and ``response`` before it returns.
        :type auth: function(dxf_obj, response)

        :param insecure: Use HTTP instead of HTTPS (which is the default) when connecting to the registry.
        :type insecure: bool

        :param auth_host: Host to use for token authentication. If set, overrides host returned by then registry.
        :type auth_host: str

        :param tlsverify: When set to False, do not verify TLS certificate. When pointed to a `<ca bundle>.crt` file use this for TLS verification. See `requests.verify <http://docs.python-requests.org/en/latest/user/advanced/#ssl-cert-verification>`_ for more details.
        :type tlsverify: bool or str
        """
        super(DXF, self).__init__(host, auth, insecure, auth_host, tlsverify)
        self._repo = repo
        self._repo_path = self._get_repo_path(repo)

    def _get_repo_path(self, repo, suffix='/'):
        repo_path = ''
        if repo:
            if self._host.endswith('docker.io') and len(repo.split('/')) == 1:
                repo_path = 'library/'
            repo_path += repo + suffix
        return repo_path

    def _request(self, method, path, **kwargs):
        return self._base_request(
            method,
            urlparse.urljoin(self._repo_path, path),
            **kwargs)

    def push_blob(self,
                  filename=None,
                  progress=None,
                  data=None, digest=None,
                  check_exists=True):
        # pylint: disable=too-many-arguments
        """
        Upload a file to the registry and return its (SHA-256) hash.

        The registry is content-addressable so the file's content (aka blob)
        can be retrieved later by passing the hash to :meth:`pull_blob`.

        :param filename: File to upload.
        :type filename: str

        :param data: Data to upload if ``filename`` isn't given. The data is uploaded in chunks and you must also pass ``digest``.
        :type data: Generator or iterator

        :param digest: Hash of the data to be uploaded in ``data``, if specified.
        :type digest: str (hex-encoded SHA-256, prefixed by ``sha256:``)

        :param progress: Optional function to call as the upload progresses. The function will be called with the hash of the file's content (or ``digest``), the blob just read from the file (or chunk from ``data``) and if ``filename`` is specified the total size of the file.
        :type progress: function(dgst, chunk, size)

        :param check_exists: Whether to check if a blob with the same hash already exists in the registry. If so, it won't be uploaded again.
        :type check_exists: bool

        :rtype: str
        :returns: Hash of file's content.
        """
        if filename is None:
            dgst = digest
        else:
            dgst = hash_file(filename)
        if check_exists:
            try:
                self._request('head', 'blobs/' + dgst)
                return dgst
            except requests.exceptions.HTTPError as ex:
                # pylint: disable=no-member
                if ex.response.status_code != requests.codes.not_found:
                    raise
        r = self._request('post', 'blobs/uploads/')
        upload_url = r.headers['Location']
        url_parts = list(urlparse.urlparse(upload_url))
        query = urlparse.parse_qs(url_parts[4])
        query.update({'digest': dgst})
        url_parts[4] = urlencode(query, True)
        url_parts[0] = 'http' if self._insecure else 'https'
        upload_url = urlparse.urlunparse(url_parts)
        if filename is None:
            data = _ReportingChunks(dgst, data, progress) if progress else data
            self._base_request('put', upload_url, data=data)
        else:
            with open(filename, 'rb') as f:
                data = _ReportingFile(dgst, f, progress) if progress else f
                self._base_request('put', upload_url, data=data)
        return dgst

    # pylint: disable=no-self-use
    def pull_blob(self, digest, size=False, chunk_size=None):
        """
        Download a blob from the registry given the hash of its content.

        :param digest: Hash of the blob's content (prefixed by ``sha256:``).
        :type digest: str

        :param size: Whether to return the size of the blob too.
        :type size: bool

        :param chunk_size: Number of bytes to download at a time. Defaults to 8192.
        :type chunk_size: int

        :rtype: iterator
        :returns: If ``size`` is falsey, a byte string iterator over the blob's content. If ``size`` is truthy, a tuple containing the iterator and the blob's size.
        """
        if chunk_size is None:
            chunk_size = 8192
        r = self._request('get', 'blobs/' + digest, stream=True)
        class Chunks(object):
            # pylint: disable=too-few-public-methods
            def __iter__(self):
                sha256 = hashlib.sha256()
                for chunk in r.iter_content(chunk_size):
                    sha256.update(chunk)
                    yield chunk
                dgst = 'sha256:' + sha256.hexdigest()
                if dgst != digest:
                    raise exceptions.DXFDigestMismatchError(dgst, digest)
        return (Chunks(), long(r.headers['content-length'])) if size else Chunks()

    def blob_size(self, digest):
        """
        Return the size of a blob in the registry given the hash of its content.

        :param digest: Hash of the blob's content (prefixed by ``sha256:``).
        :type digest: str

        :rtype: long
        :returns: Size of the blob in bytes.
        """
        r = self._request('head', 'blobs/' + digest)
        return long(r.headers['content-length'])

    def mount_blob(self, repo, digest):
        """
        Mount a blob from another repository in the registry.

        :param repo: Repository containing the existing blob.
        :type repo: str

        :param digest: Hash of the existing blob's content (prefixed by ``sha256:``).
        :type digest: str

        :rtype: str
        :returns: Hash of blob's content.
        """
        r = self._request('post', 'blobs/uploads/?' + urlencode({
            'mount': digest,
            'from': self._get_repo_path(repo, suffix='')
        }))
        # pylint: disable=no-member
        if r.status_code != requests.codes.created:
            raise exceptions.DXFMountFailed()
        return r.headers.get('Docker-Content-Digest')

    def del_blob(self, digest):
        """
        Delete a blob from the registry given the hash of its content.

        :param digest: Hash of the blob's content (prefixed by ``sha256:``).
        :type digest: str
        """
        self._request('delete', 'blobs/' + digest)

    # For dtuf; highly unlikely anyone else will want this
    def make_manifest(self, *digests):
        layers = [{
            'mediaType': 'application/octet-stream',
            'size': self.blob_size(dgst),
            'digest': dgst
        } for dgst in digests]
        return json.dumps({
            'schemaVersion': 2,
            'mediaType': 'application/vnd.docker.distribution.manifest.v2+json',
            # V2 Schema 2 insists on a config dependency. We're just using the
            # registry as a blob store so to save us uploading extra blobs,
            # use the first layer.
            'config': {
                'mediaType': 'application/vnd.docker.container.image.v1+json',
                'size': layers[0]['size'],
                'digest': layers[0]['digest']
            },
            'layers': layers
        }, sort_keys=True)

    def set_manifest(self, alias, manifest_json):
        """
        Give a name (alias) to a manifest.

        :param alias: Alias name
        :type alias: str

        :param manifest_json: A V2 Schema 2 manifest JSON string
        :type digests: list
        """
        self._request('put',
                      'manifests/' + alias,
                      data=manifest_json,
                      headers={'Content-Type': _schema2_mimetype})

    def set_alias(self, alias, *digests):
        # pylint: disable=too-many-locals
        """
        Give a name (alias) to a set of blobs. Each blob is specified by
        the hash of its content.

        :param alias: Alias name
        :type alias: str

        :param digests: List of blob hashes (prefixed by ``sha256:``).
        :type digests: list of strings

        :rtype: str
        :returns: The registry manifest used to define the alias. You almost definitely won't need this.
        """
        try:
            manifest_json = self.make_manifest(*digests)
            self.set_manifest(alias, manifest_json)
            return manifest_json
        except requests.exceptions.HTTPError as ex:
            # pylint: disable=no-member
            if ex.response.status_code != requests.codes.bad_request:
                raise
            manifest_json = self.make_unsigned_manifest(alias, *digests)
            signed_json = _sign_manifest(manifest_json)
            self._request('put', 'manifests/' + alias, data=signed_json)
            return signed_json

    def get_manifest_and_response(self, alias):
        """
        Request the manifest for an alias and return the manifest and the
        response.

        :param alias: Alias name.
        :type alias: str

        :rtype: tuple
        :returns: Tuple containing the manifest as a string (JSON) and the `requests.Response <http://docs.python-requests.org/en/master/api/#requests.Response>`_
        """
        r = self._request('get',
                          'manifests/' + alias,
                          headers={'Accept': _schema2_mimetype + ', ' +
                                             _schema1_mimetype})
        return r.content.decode('utf-8'), r

    def get_manifest(self, alias):
        """
        Get the manifest for an alias

        :param alias: Alias name.
        :type alias: str

        :rtype: str
        :returns: The manifest as string (JSON)
        """
        manifest, _ = self.get_manifest_and_response(alias)
        return manifest

    def _get_alias(self, alias, manifest, verify, sizes, dcd, get_digest):
        # pylint: disable=too-many-arguments
        if alias:
            manifest, r = self.get_manifest_and_response(alias)
            dcd = r.headers['docker-content-digest']

        parsed_manifest = json.loads(manifest)

        if parsed_manifest['schemaVersion'] == 1:
            # https://github.com/docker/distribution/issues/1662#issuecomment-213101772
            # "A schema1 manifest should always produce the same image id but
            # defining the steps to produce directly from the manifest is not
            # straight forward."
            if get_digest:
                raise exceptions.DXFDigestNotAvailableForSchema1()

            r = _verify_manifest(manifest,
                                 parsed_manifest,
                                 dcd,
                                 verify)

            return [(dgst, self.blob_size(dgst)) for dgst in r] if sizes else r

        if dcd:
            method, expected_dgst = split_digest(dcd)
            hasher = hashlib.new(method)
            hasher.update(r.content)
            dgst = hasher.hexdigest()
            if dgst != expected_dgst:
                raise exceptions.DXFDigestMismatchError(
                    method + ':' + dgst,
                    method + ':' + expected_dgst)

        if get_digest:
            dgst = parsed_manifest['config']['digest']
            split_digest(dgst)
            return dgst

        r = []
        for layer in parsed_manifest['layers']:
            dgst = layer['digest']
            split_digest(dgst)
            r.append((dgst, layer['size']) if sizes else dgst)
        return r

    def get_alias(self,
                  alias=None,
                  manifest=None,
                  verify=True,
                  sizes=False,
                  dcd=None):
        # pylint: disable=too-many-arguments
        """
        Get the blob hashes assigned to an alias.

        :param alias: Alias name. You almost definitely will only need to pass this argument.
        :type alias: str

        :param manifest: If you previously obtained a manifest, specify it here instead of ``alias``. You almost definitely won't need to do this.
        :type manifest: str

        :param verify: (v1 schema only) Whether to verify the integrity of the alias definition in the registry itself. You almost definitely won't need to change this from the default (``True``).
        :type verify: bool

        :param sizes: Whether to return sizes of the blobs along with their hashes
        :type sizes: bool

        :param dcd: (if ``manifest`` is specified) The Docker-Content-Digest header returned when getting the manifest. If present, this is checked against the manifest.
        :type dcd: str

        :rtype: list
        :returns: If ``sizes`` is falsey, a list of blob hashes (strings) which are assigned to the alias. If ``sizes`` is truthy, a list of (hash,size) tuples for each blob.
        """
        return self._get_alias(alias, manifest, verify, sizes, dcd, False)

    def get_digest(self,
                   alias=None,
                   manifest=None,
                   verify=True,
                   dcd=None):
        """
        (v2 schema only) Get the hash of an alias's configuration blob.

        For an alias created using ``dxf``, this is the hash of the first blob
        assigned to the alias.

        For a Docker image tag, this is the same as
        ``docker inspect alias --format='{{.Id}}'``.

        :param alias: Alias name. You almost definitely will only need to pass this argument.
        :type alias: str

        :param manifest: If you previously obtained a manifest, specify it here instead of ``alias``. You almost definitely won't need to do this.
        :type manifest: str

        :param verify: (v1 schema only) Whether to verify the integrity of the alias definition in the registry itself. You almost definitely won't need to change this from the default (``True``).
        :type verify: bool

        :param dcd: (if ``manifest`` is specified) The Docker-Content-Digest header returned when getting the manifest. If present, this is checked against the manifest.
        :type dcd: str

        :rtype: str
        :returns: Hash of the alias's configuration blob.
        """
        return self._get_alias(alias, manifest, verify, False, dcd, True)

    def _get_dcd(self, alias):
        """
        Get the Docker-Content-Digest header for an alias.

        :param alias: Alias name.
        :type alias: str

        :rtype: str
        :returns: DCD header for the alias.
        """
        # https://docs.docker.com/registry/spec/api/#deleting-an-image
        # Note When deleting a manifest from a registry version 2.3 or later,
        # the following header must be used when HEAD or GET-ing the manifest
        # to obtain the correct digest to delete:
        # Accept: application/vnd.docker.distribution.manifest.v2+json
        return self._request(
            'head',
            'manifests/{}'.format(alias),
            headers={'Accept': _schema2_mimetype},
        ).headers.get('Docker-Content-Digest')

    def del_alias(self, alias):
        """
        Delete an alias from the registry. The blobs it points to won't be deleted. Use :meth:`del_blob` for that.

        .. Note::
           On private registry, garbage collection might need to be run manually; see:
           https://docs.docker.com/registry/garbage-collection/

        :param alias: Alias name.
        :type alias: str

        :rtype: list
        :returns: A list of blob hashes (strings) which were assigned to the alias.
        """
        dcd = self._get_dcd(alias)
        dgsts = self.get_alias(alias)
        self._request('delete', 'manifests/{}'.format(dcd))
        return dgsts

    def list_aliases(self, batch_size=None, iterate=False):
        """
        List all aliases defined in the repository.

        :param batch_size: Number of alias names to ask the server for at a time.
        :type batch_size: int

        :param iterate: Whether to return iterator over the names or a list of all the names.
        :type iterate: bool

        :rtype: list or iterator of strings
        :returns: Alias names.
        """
        it = PaginatingResponse(self, '_request',
                                'tags/list', 'tags',
                                params={'n': batch_size})
        return it if iterate else list(it)

    def api_version_check(self):
        """
        Performs API version check

        :rtype: tuple
        :returns: verson check response as a string (JSON) and requests.Response
        """
        r = self._base_request('get', '')
        return r.content.decode('utf-8'), r

    @classmethod
    def from_base(cls, base, repo):
        """
        Create a :class:`DXF` object which uses the same host, settings and
        session as an existing :class:`DXFBase` object.

        :param base: Existing :class:`DXFBase` object.
        :type base: :class:`DXFBase`

        :param repo: Name of the repository to access on the registry. Typically this is of the form ``username/reponame`` but for your own registries you don't actually have to stick to that.
        :type repo: str

        :returns: :class:`DXF` object which shares configuration and session with ``base`` but which can also be used to operate on the ``repo`` repository.
        :rtype: :class:`DXF`
        """
        # pylint: disable=protected-access
        r = cls(base._host, repo, base._auth, base._insecure, base._auth_host, base._tlsverify)
        r._token = base._token
        r._headers = base._headers
        r._sessions = [base._sessions[0]]
        return r

# v1 schema support functions below

    def make_unsigned_manifest(self, alias, *digests):
        return json.dumps({
            'schemaVersion': 1,
            'name': self._repo,
            'tag': alias,
            'fsLayers': [{'blobSum': dgst} for dgst in digests],
            'history': [{'v1Compatibility': '{}'} for dgst in digests]
        }, sort_keys=True)

_schema1_mimetype = 'application/vnd.docker.distribution.manifest.v1+json'

def _urlsafe_b64encode(s):
    return base64.urlsafe_b64encode(_to_bytes_2and3(s)).rstrip(b'=').decode('utf-8')

def _pad64(s):
    return s + b'=' * (-len(s) % 4)

def _urlsafe_b64decode(s):
    return base64.urlsafe_b64decode(_pad64(_to_bytes_2and3(s)))

def _import_key(expkey):
    if expkey['kty'] != 'EC':
        raise exceptions.DXFUnexpectedKeyTypeError(expkey['kty'], 'EC')
    if expkey['crv'] != 'P-256':
        raise exceptions.DXFUnexpectedKeyTypeError(expkey['crv'], 'P-256')
    return jwk.JWK(kty='EC', crv='P-256', x=expkey['x'], y=expkey['y'])

def _sign_manifest(manifest_json):
    format_length = manifest_json.rfind('}')
    format_tail = manifest_json[format_length:]
    key = jwk.JWK.generate(kty='EC', crv='P-256')
    jwstoken = jws.JWS(manifest_json.encode('utf-8'))
    jkey = json.loads(key.export_public())
    # Docker expects 32 bytes for x and y
    jkey['x'] = _urlsafe_b64encode(_urlsafe_b64decode(jkey['x']).rjust(32, b'\0'))
    jkey['y'] = _urlsafe_b64encode(_urlsafe_b64decode(jkey['y']).rjust(32, b'\0'))
    jwstoken.add_signature(key, None, {
        'formatLength': format_length,
        'formatTail': _urlsafe_b64encode(format_tail)
    }, {
        'jwk': jkey,
        'alg': 'ES256'
    })
    return manifest_json[:format_length] + \
           ', "signatures": [' + jwstoken.serialize() + ']' + \
           format_tail

def _verify_manifest(content,
                     manifest,
                     content_digest=None,
                     verify=True):
    # pylint: disable=too-many-locals,too-many-branches

    # Adapted from https://github.com/joyent/node-docker-registry-client

    if verify or ('signatures' in manifest):
        signatures = []
        for sig in manifest['signatures']:
            protected64 = sig['protected']
            protected = _urlsafe_b64decode(protected64).decode('utf-8')
            protected_header = json.loads(protected)

            format_length = protected_header['formatLength']
            format_tail64 = protected_header['formatTail']
            format_tail = _urlsafe_b64decode(format_tail64).decode('utf-8')

            alg = sig['header']['alg']
            if alg.lower() == 'none':
                raise exceptions.DXFDisallowedSignatureAlgorithmError('none')
            if sig['header'].get('chain'):
                raise exceptions.DXFSignatureChainNotImplementedError()

            signatures.append({
                'alg': alg,
                'signature': sig['signature'],
                'protected64': protected64,
                'key': _import_key(sig['header']['jwk']),
                'format_length': format_length,
                'format_tail': format_tail
            })

        payload = content[:signatures[0]['format_length']] + \
                  signatures[0]['format_tail']
        payload64 = _urlsafe_b64encode(payload)
    else:
        payload = content

    if content_digest:
        method, expected_dgst = split_digest(content_digest)
        hasher = hashlib.new(method)
        hasher.update(payload.encode('utf-8'))
        dgst = hasher.hexdigest()
        if dgst != expected_dgst:
            raise exceptions.DXFDigestMismatchError(
                method + ':' + dgst,
                method + ':' + expected_dgst)

    if verify:
        for sig in signatures:
            jwstoken = jws.JWS()
            jwstoken.deserialize(json.dumps({
                'payload': payload64,
                'protected': sig['protected64'],
                'signature': sig['signature']
            }), sig['key'], sig['alg'])

    dgsts = []
    for layer in manifest['fsLayers']:
        dgst = layer['blobSum']
        split_digest(dgst)
        dgsts.append(dgst)
    return dgsts



import io
import os
from setuptools import setup

def read(name):
    file_path = os.path.join(os.path.dirname(__file__), name)
    return io.open(file_path, encoding='utf8').read()

setup(
    name='python-dxf',
    version='7.8.0',
    description="Package for accessing a Docker v2 registry",
    long_description=read('README.rst'),
    keywords='docker registry',
    author='David Halls',
    author_email='dave@davedoesdev.com',
    url='https://github.com/davedoesdev/dxf',
    license='MIT',
    packages=['dxf'],
    entry_points={'console_scripts': ['dxf=dxf.main:main']},
    install_requires=['www-authenticate>=0.9.2',
                      'requests>=2.18.4',
                      "jwcrypto>=0.4.2,<1.0.0; python_version < '3.0'",
                      "jwcrypto>=1.0.0; python_version >= '3.0'",
                      'tqdm>=4.19.4']
)



#pylint: disable=wrong-import-position,wrong-import-order,superfluous-parens
import os
import argparse
import sys
import traceback
import errno
import tqdm
import dxf
import dxf.exceptions
import requests.exceptions

_choices = ['auth',
            'push-blob',
            'pull-blob',
            'blob-size',
            'mount-blob',
            'del-blob',
            'set-alias',
            'get-alias',
            'del-alias',
            'get-digest',
            'list-aliases',
            'list-repos']

_parser = argparse.ArgumentParser()
_subparsers = _parser.add_subparsers(dest='op')
for c in _choices:
    sp = _subparsers.add_parser(c)
    if c != 'list-repos':
        sp.add_argument("repo")
        sp.add_argument('args', nargs='*')

def _flatten(l):
    return [item for sublist in l for item in sublist]

# pylint: disable=too-many-statements
def doit(args, environ):
    dxf_progress = environ.get('DXF_PROGRESS')
    if dxf_progress == '1' or (dxf_progress != '0' and sys.stderr.isatty()):
        bars = {}
        def progress(dgst, chunk, size):
            if dgst not in bars:
                bars[dgst] = tqdm.tqdm(desc=dgst[0:8],
                                       total=size,
                                       leave=True)
            if chunk:
                bars[dgst].update(len(chunk))
            if bars[dgst].n >= bars[dgst].total:
                bars[dgst].close()
                del bars[dgst]
    else:
        progress = None

    dxf_skiptlsverify = environ.get('DXF_SKIPTLSVERIFY')
    if dxf_skiptlsverify == '1':
        dxf_tlsverify = False
    else:
        dxf_tlsverify = environ.get('DXF_TLSVERIFY', True)

    def auth(dxf_obj, response):
        # pylint: disable=redefined-outer-name
        username = environ.get('DXF_USERNAME')
        password = environ.get('DXF_PASSWORD')
        authorization = environ.get('DXF_AUTHORIZATION')
        dxf_obj.authenticate(username, password,
                             response=response,
                             authorization=authorization)

    args = _parser.parse_args(args)
    if args.op != 'list-repos':
        dxf_obj = dxf.DXF(environ['DXF_HOST'],
                          args.repo,
                          auth,
                          environ.get('DXF_INSECURE') == '1',
                          environ.get('DXF_AUTH_HOST'),
                          tlsverify=dxf_tlsverify)
    else:
        dxf_obj = dxf.DXFBase(environ['DXF_HOST'],
                              auth,
                              environ.get('DXF_INSECURE') == '1',
                              environ.get('DXF_AUTH_HOST'),
                              tlsverify=dxf_tlsverify)

    def _doit():
        # pylint: disable=too-many-branches
        if args.op == "auth":
            username = environ.get('DXF_USERNAME')
            password = environ.get('DXF_PASSWORD')
            authorization = environ.get('DXF_AUTHORIZATION')
            token = dxf_obj.authenticate(username, password,
                                         actions=args.args,
                                         authorization=authorization)
            if token:
                print(token)
            return

        token = environ.get('DXF_TOKEN')
        if token:
            dxf_obj.token = token

        if args.op == "push-blob":
            if len(args.args) < 1:
                _parser.error('too few arguments')
            if len(args.args) > 2:
                _parser.error('too many arguments')
            if len(args.args) == 2 and not args.args[1].startswith('@'):
                _parser.error('invalid alias')
            dgst = dxf_obj.push_blob(args.args[0], progress)
            if len(args.args) == 2:
                dxf_obj.set_alias(args.args[1][1:], dgst)
            print(dgst)

        elif args.op == "pull-blob":
            _stdout = getattr(sys.stdout, 'buffer', sys.stdout)
            if args.args:
                dgsts = _flatten([dxf_obj.get_alias(name[1:])
                                  if name.startswith('@') else [name]
                                  for name in args.args])
            else:
                dgsts = dxf_obj.get_alias(manifest=sys.stdin.read())
            for dgst in dgsts:
                it, size = dxf_obj.pull_blob(
                    dgst, size=True, chunk_size=environ.get('DXF_CHUNK_SIZE'))
                if environ.get('DXF_BLOB_INFO') == '1':
                    print(dgst + ' ' + str(size))
                if progress:
                    progress(dgst, b'', size)
                for chunk in it:
                    if progress:
                        progress(dgst, chunk, size)
                    _stdout.write(chunk)

        elif args.op == 'blob-size':
            if args.args:
                sizes = [dxf_obj.get_alias(name[1:], sizes=True)
                         if name.startswith('@') else
                         [(name, dxf_obj.blob_size(name))]
                         for name in args.args]
            else:
                sizes = [dxf_obj.get_alias(manifest=sys.stdin.read(),
                                           sizes=True)]
            for tuples in sizes:
                print(sum([size for _, size in tuples]))

        elif args.op == 'mount-blob':
            if len(args.args) < 2:
                _parser.error('too few arguments')
            if len(args.args) > 3:
                _parser.error('too many arguments')
            if len(args.args) == 3 and not args.args[2].startswith('@'):
                _parser.error('invalid alias')
            dgst = dxf_obj.mount_blob(args.args[0], args.args[1])
            if len(args.args) == 3:
                dxf_obj.set_alias(args.args[2][1:], dgst)
            print(dgst)

        elif args.op == 'del-blob':
            if args.args:
                dgsts = _flatten([dxf_obj.del_alias(name[1:])
                                  if name.startswith('@') else [name]
                                  for name in args.args])
            else:
                dgsts = dxf_obj.get_alias(manifest=sys.stdin.read())
            for dgst in dgsts:
                dxf_obj.del_blob(dgst)

        elif args.op == "set-alias":
            if len(args.args) < 2:
                _parser.error('too few arguments')
            dgsts = [dxf.hash_file(dgst) if os.sep in dgst else dgst
                     for dgst in args.args[1:]]
            sys.stdout.write(dxf_obj.set_alias(args.args[0], *dgsts))

        elif args.op == "get-alias":
            if args.args:
                dgsts = _flatten([dxf_obj.get_alias(name) for name in args.args])
            else:
                dgsts = dxf_obj.get_alias(manifest=sys.stdin.read())
            for dgst in dgsts:
                print(dgst)

        elif args.op == "del-alias":
            for name in args.args:
                for dgst in dxf_obj.del_alias(name):
                    print(dgst)

        elif args.op == 'get-digest':
            if args.args:
                dgsts = [dxf_obj.get_digest(name) for name in args.args]
            else:
                dgsts = [dxf_obj.get_digest(manifest=sys.stdin.read())]
            for dgst in dgsts:
                print(dgst)

        elif args.op == 'list-aliases':
            if args.args:
                _parser.error('too many arguments')
            for name in dxf_obj.list_aliases(iterate=True):
                print(name)

        elif args.op == 'list-repos':
            for name in dxf_obj.list_repos(iterate=True):
                print(name)

    try:
        _doit()
        return 0
    except dxf.exceptions.DXFUnauthorizedError:
        traceback.print_exc()
        return errno.EACCES
    except requests.exceptions.HTTPError as ex:
        # pylint: disable=no-member
        if ex.response.status_code == requests.codes.not_found:
            traceback.print_exc()
            return errno.ENOENT
        raise

def main():
    sys.exit(doit(sys.argv[1:], os.environ))
